/*
 *
 * nucleo/core/carbon/cReactiveEngine.cxx --
 *
 * Copyright (C) Nicolas Roussel
 *
 * See the file LICENSE for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <nucleo/config.H>

#include <nucleo/core/carbon/cReactiveEngine.H>
#include <nucleo/gl/window/glWindow.H>

#include <Carbon/Carbon.h>

#include <iostream>

namespace nucleo {

#define NUCLEO_USES_ITS_OWN_EVENT_QUEUE 1

  static const long cReactiveEngine_event_class = 'nucE' ;
  static const long cReactiveEngine_event_kind = 'noti' ;

  // ----------------------------------------------------------

  cReactiveEngine::cReactiveEngine(void) {
    EventTypeSpec ets ;
    ets.eventClass = cReactiveEngine_event_class ;
    ets.eventKind = cReactiveEngine_event_kind ;
    _eventQueued = false ;
    _eQueue = GetCurrentEventQueue() ;

    InstallEventHandler(GetApplicationEventTarget(),
				    NewEventHandlerUPP(_handleEvent),
				    1, &ets,
				    (void *)this,
				    NULL /* outRef */) ;
  }

  cReactiveEngine::~cReactiveEngine(void) {
  }

  // ----------------------------------------------------------

  OSStatus
  cReactiveEngine::_handleEvent(EventHandlerCallRef nextHandler,
						  EventRef theEvent, void* userData) {
    cReactiveEngine *myself = (cReactiveEngine*)userData ;

#if NUCLEO_USES_ITS_OWN_EVENT_QUEUE
    unsigned int nb2notify = myself->_objectsToNotify.size() ;
    // std::cerr << "cReactiveEngine::_handleEvent: handling " << nb2notify << " notifications" << std::endl ;
    for (unsigned int i=0; i<nb2notify; ++i) {
	 ReactivePair pair = myself->_objectsToNotify.front() ; 
	 myself->_objectsToNotify.pop() ;
	 if (ReactiveObject::isAlive(pair.first)) {
	   // std::cerr << "cReactiveEngine: calling react on " << pair.first << std::endl ;
	   pair.first->react(pair.second) ;
	 } else {
	   myself->detachFrom(pair.first, pair.second) ;
	 }
	 myself->incPendingNotifications(pair.second, -1) ;
    }

    nb2notify = myself->_objectsToNotify.size() ;
    if (nb2notify) {
	 // std::cerr << "cReactiveEngine::_handleEvent: still " << nb2notify << " notifications to deal with" << std::endl ;
	 EventRef event = 0 ;
	 CreateEvent(0,
			   cReactiveEngine_event_class, cReactiveEngine_event_kind,
			   0, kEventAttributeUserEvent, &event) ;
	 PostEventToQueue(myself->_eQueue, event, kEventPriorityStandard) ;
	 ReleaseEvent(event) ;
	 myself->_eventQueued = true ;
    } else
	 myself->_eventQueued = false ;

    return noErr ;
#else
    ReactivePair *pair = 0 ;
    OSStatus result = GetEventParameter(theEvent, kEventParamUserData, typeVoidPtr, NULL,
								sizeof(pair), NULL, &pair) ;
    if (result==noErr && pair) {
	 if (ReactiveObject::isAlive(pair->first)) {
	   // std::cerr << pair->first << " reacting to " << pair->second << std::endl ;
	   pair->first->react(pair->second) ;
	 } else {
	   // std::cerr << "cReactiveEngine: ReactiveObject " << pair->first << " is dead..." << std::endl ;
	   myself->detachFrom(pair->first, pair->second) ;
	 }
	 myself->incPendingNotifications(pair->second, -1) ;

	 delete pair ;
	 return noErr ;
    }
    return eventNotHandledErr ;
#endif
  }

  void
  cReactiveEngine::notify(ReactiveObject *notified, Observable *notifier) {
    incPendingNotifications(notifier, +1) ;   

    EventRef event = 0 ;
#if NUCLEO_USES_ITS_OWN_EVENT_QUEUE
    if (!_eventQueued)
	 CreateEvent(0,
			   cReactiveEngine_event_class, cReactiveEngine_event_kind,
			   0, kEventAttributeUserEvent, &event) ;
    _objectsToNotify.push(ReactivePair(notified,notifier)) ;
#else
    CreateEvent(0,
			 cReactiveEngine_event_class, cReactiveEngine_event_kind,
			 0, kEventAttributeUserEvent, &event) ;
    ReactivePair *pair = new ReactivePair(notified,notifier) ;
    SetEventParameter(event,
				  kEventParamUserData, typeVoidPtr,
				  sizeof(pair), (void *)&pair) ;
#endif

    if (event) {
	 PostEventToQueue(_eQueue, event, kEventPriorityStandard) ;
	 ReleaseEvent(event) ;
    }

#if 0
    UInt32 ne = GetNumEventsInQueue(eQueue) ;
    if (ne>1) std::cerr << ne << " events queued" << std::endl ;
#endif
  }

  // ----------------------------------------------------------

  int
  cReactiveEngine::step(long milliseconds) {
    EventRef theEvent ;
    EventTargetRef theTarget = GetEventDispatcherTarget() ;
    EventTimeout carbon_timeout = (milliseconds==-1) ? kEventDurationForever : milliseconds*kEventDurationMillisecond ;
    if (ReceiveNextEvent(0, NULL, carbon_timeout, true, &theEvent)==noErr) {
	 SendEventToEventTarget(theEvent, theTarget) ;
	 ReleaseEvent(theEvent) ;
	 return 1 ;
    }
    return 0 ;
  }

  void
  cReactiveEngine::run(void) {
    RunApplicationEventLoop() ;
  }

  void
  cReactiveEngine::stop(void) {
    QuitApplicationEventLoop() ;
  }

}
